#!/usr/bin/env python
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Quick analysis tool for the compreffor. Compresses a directory of
fonts and writes a CSV containing various stats about the compression.

Usage (command line):
>>> ./analyzer.py /path/to/font/dir
...

A CSV will be written to the current working directory with
the name data.csv.
"""

import csv
import os
import time
from compreffor import cxxCompressor

from fontTools import subset
from fontTools.ttLib import TTFont

def sum_subrs(font):
   td = font['CFF '].cff.topDictIndex[0]
   ans = len(td.GlobalSubrs)
   try:
      ans += sum(len(fd.Private.Subrs) if hasattr(fd.Private, 'Subrs') else 0 for fd in td.FDArray)
   except AttributeError:
      pass
   return ans

if __name__ == '__main__':
   names = []
   times = []
   orig_sizes = []
   full_sizes = []
   compressed_sizes = []
   nsubrs = []
   nsubrs_orig = []

   for root, dirs, files in os.walk(os.argv[1]):
      for filename in files:
         if os.path.splitext(filename)[1] == '.otf':
            fname = os.path.join(root, filename)
            print "Handling %s" % filename

            # decompress
            print("\tDecompressing...")
            font = TTFont(fname)
            orig_subrs = sum_subrs(font)
            orig_size = os.path.getsize(fname)

            options = subset.Options()
            options.decompress = True
            subsetter = subset.Subsetter(options=options)
            subsetter.populate(glyphs=font.getGlyphOrder())
            subsetter.subset(font)
            name_parts = os.path.splitext(fname)
            new_fname = name_parts[0] + '-decomp' + name_parts[1]
            font.save(new_fname)
            full_size = os.path.getsize(new_fname)

            print("\tSubroutinizing...")
            print("----")
            start_time = time.time()
            cxxCompressor.main(filename=new_fname, verbose=True)
            times.append(time.time() - start_time)
            print("----")

            print("\tTabulating results...")
            comp_fname = name_parts[0] + '-decomp.compressed' + name_parts[1]
            comp_subrs = sum_subrs(TTFont(comp_fname))
            comp_size = os.path.getsize(comp_fname)

            orig_sizes.append(orig_size)
            full_sizes.append(full_size)
            compressed_sizes.append(comp_size)
            names.append(filename)
            nsubrs_orig.append(orig_subrs)
            nsubrs.append(comp_subrs)

   with open('data.csv', 'w') as csvf:
      nwriter = csv.writer(csvf)
      nwriter.writerow(['Name', 'Time to compress', 'Original Size',
                        'Expanded Size', 'Compressed Size',
                        'Original # of Subrs', 'Compressed # of Subrs'])
      nwriter.writerows(zip(names, times, orig_sizes, full_sizes, 
                            compressed_sizes, nsubrs_orig, nsubrs))



import os
import argparse
import logging

import compreffor
from compreffor.pyCompressor import human_size
from fontTools.ttLib import TTFont
from fontTools.misc.loggingTools import configLogger


def parse_arguments(args=None):
    parser = argparse.ArgumentParser(
        prog='compreffor',
        description="FontTools Compreffor will take a CFF-flavored OpenType font "
                    "and automatically detect repeated routines and generate "
                    "subroutines to minimize the disk space needed to represent "
                    "a font.")
    parser.add_argument("infile", metavar="INPUT",
                        help="path to the input font file")
    parser.add_argument("outfile", nargs="?", metavar="OUTPUT",
                        help="path to the compressed file (default: "
                        "*.compressed.otf)")
    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help="print more messages to stdout; use it multiple "
                        "times to increase the level of verbosity")
    parser.add_argument("-c", "--check", action='store_true',
                        help="verify that the outputted font is valid and "
                             "functionally equivalent to the input")
    parser.add_argument("-d", "--decompress", action="store_true",
                        help="decompress source before compressing (necessary if "
                             "there are subroutines in the source)")
    parser.add_argument("-n", "--nrounds", type=int,
                        help="the number of iterations to run the algorithm"
                             " (default: 4)")
    parser.add_argument("-m", "--max-subrs", type=int,
                        help="limit to the number of subroutines per INDEX "
                        " (default: 65533)")
    parser.add_argument('--generate-cff', action='store_true',
                        help="Also save binary CFF table data as {INPUT}.cff")
    parser.add_argument('--py', dest="method_python", action='store_true',
                        help="Use pure Python method, instead of C++ extension")
    py_meth_group = parser.add_argument_group("options for pure Python method")
    py_meth_group.add_argument("--chunk-ratio", type=float,
                               help="specify the percentage size of the "
                                    "job chunks used for parallel processing "
                                    "(0 < float <= 1; default: 0.1)")
    py_meth_group.add_argument("-p", "--processes", type=int,
                               help="specify number  of concurrent processes to "
                               "run. Use value 1 to perform operation serially "
                               "(default: 12)")

    options = parser.parse_args(args)
    kwargs = vars(options)

    if options.method_python:
        if options.processes is not None and options.processes < 1:
            parser.error('argument --processes expects positive integer > 0')
        if (options.chunk_ratio is not None
                and not (0 < options.chunk_ratio <= 1)):
            parser.error('argument --chunk-ratio expects float number 0 < n <= 1')
    else:
        for attr in ('chunk_ratio', 'processes'):
            if getattr(options, attr):
                opt = attr.replace('_', '-')
                parser.error('argument --%s can only be used with --py (pure '
                             'Python) method' % opt)
            else:
                del kwargs[attr]
    if options.outfile is None:
        options.outfile = "%s.compressed%s" % os.path.splitext(options.infile)

    return kwargs


def main(args=None):
    log = compreffor.log
    timer = compreffor.timer
    timer.reset()

    options = parse_arguments(args)

    # consume kwargs that are not passed on to 'compress' function
    infile = options.pop('infile')
    outfile = options.pop('outfile')
    decompress = options.pop('decompress')
    generate_cff = options.pop('generate_cff')
    check = options.pop('check')
    verbose = options.pop('verbose')

    if verbose == 1:
        level = logging.INFO
    elif verbose > 1:
        level = logging.DEBUG
    else:
        level = logging.WARNING

    configLogger(logger=log, level=level)

    orig_size = os.path.getsize(infile)

    font = TTFont(infile)

    if decompress:
        log.info("Decompressing font with FontTools Subsetter")
        with timer("decompress the font"):
            compreffor.decompress(font)

    log.info("Compressing font through %s Compreffor",
             "pure-Python" if options['method_python'] else "C++")

    compreffor.compress(font, **options)

    with timer("compile and save compressed font"):
        font.save(outfile)

    if generate_cff:
        cff_file = os.path.splitext(outfile)[0] + ".cff"
        with open(cff_file, 'wb') as f:
            font['CFF '].cff.compile(f, None)
        log.info("Saved CFF data to '%s'" % os.path.basename(cff_file))

    if check:
        log.info("Checking compression integrity and call depth")
        passed = compreffor.check(infile, outfile)
        if not passed:
            return 1

    comp_size = os.path.getsize(outfile)
    log.info("Compressed to '%s' -- saved %s" %
             (os.path.basename(outfile), human_size(orig_size - comp_size)))
    log.debug("Total time: %gs", timer.time())


if __name__ == "__main__":
    main()



#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import random
from compreffor import pyCompressor
from compreffor.test.dummy import DummyGlyphSet


class TestCffCompressor(unittest.TestCase):

    def setUp(self):
        self.glyph_set = DummyGlyphSet({'a': (0, 1, 20, 21, 22, 2), 'b': (7, 0, 1, 20, 21, 22, 2), 'c': (0, 1, 20, 21, 22, 9, 3, 17)})
        self.sf = pyCompressor.SubstringFinder(self.glyph_set)

        self.short_sf = pyCompressor.SubstringFinder(DummyGlyphSet({'a': (1, 2, 3), 'b': (8, 1, 4)}))

        self.rand_gs = DummyGlyphSet()
        num_glyphs = random.randint(5, 20)
        for i in range(num_glyphs):
            length = random.randint(2, 30)
            self.rand_gs[i] = tuple(random.randint(0, 100) for _ in range(length))
        self.random_sf = pyCompressor.SubstringFinder(DummyGlyphSet(self.rand_gs))

        length = 3
        locations = [(0, 0), (1, 4)]
        charstrings = [(348, 374, 'rmoveto', 'endchar'), (123, -206, -140, 'hlineto', 348, 374, 'rmoveto', 'endchar')]

        self.cand_subr = pyCompressor.CandidateSubr(length, locations[0], 2, charstrings)

        self.empty_compreffor = pyCompressor.Compreffor(None, test_mode=True)

    def test_iterative_encode(self):
        """Test iterative_encode function"""

        ans = self.empty_compreffor.iterative_encode(self.glyph_set)
        self.assertIsInstance(ans, dict)

        encs = ans["glyph_encodings"]

        expected_subr_length = 5 # subr is (0, 1, 20, 21, 22)

        for glyph_enc in encs.values():
            self.assertTrue(any(cs[1].length == expected_subr_length for cs in glyph_enc))


    def test_get_substrings_all(self):
        """Test get_substrings without restrictions"""

        ans = [s.value() for s in self.sf.get_substrings(0, False)]

        expected_values = [(0, 1, 2, 3, 4, 5), (0, 1, 2, 3, 4), (1, 2, 3, 4, 5), (1, 2, 3, 4), \
                            (2, 3, 4, 5), (2, 3, 4), (3, 4, 5), (3, 4), (4, 5), (4,), (5,)]

        self.assertEqual(ans, expected_values)

    def test_get_substrings_standard(self):
        """Check to make sure all substrings have freq >= 2 and positive savings"""
        ans = self.sf.get_substrings()

        for substr in ans:
            self.assertTrue(substr.freq >= 2)
            self.assertTrue(substr.subr_saving() > 0)

    def test_get_suffixes(self):
        """Test the results of suffix array construction."""

        ans = self.short_sf.get_suffixes()

        self.assertEqual(ans, [(0, 0), (1, 1), (0, 1), (0, 2), (1, 0), (1, 2)])

    def test_get_suffixes_random(self):
        """Check suffix array invariants on random input"""

        ans = self.random_sf.get_suffixes()

        # check there are the right number of suffixes
        expected_num = sum([len(chstring) for chstring in self.rand_gs.values()])
        actual_num = len(ans)
        self.assertEqual(actual_num, expected_num)

        # check that the order is correct
        last_glidx, last_tidx = ans[0]
        last = self.random_sf.data[last_glidx][last_tidx:]
        for glyph_idx, tok_idx in ans[1:]:
            current = self.random_sf.data[glyph_idx][tok_idx:]
            self.assertTrue(last <= current)

    def test_get_lcp(self):
        """Test the lcp array generation"""

        expected = [0, 6, 5, 0, 5, 4, 0, 4, 3, 0, 3, 2, 0, 2, 1, 0, 1, 0, 0, 0, 0]

        self.assertEqual(self.sf.get_lcp(), expected)

    def test_human_size(self):
        """Test the human_size function for various numbers of bytes"""

        human_size = pyCompressor.human_size

        self.assertEqual(human_size(2), "2.0 bytes")
        self.assertEqual(human_size(2050), "2.0 KB")
        self.assertEqual(human_size(3565158), "3.4 MB")
        self.assertEqual(human_size(6120328397), "5.7 GB")

    def test_collect_lsubrs_called_from(self):
        """Test collecting local subrs called from any global subrs"""

        g1 = pyCompressor.CandidateSubr(3, (0, 10))
        g1._global = True
        g2 = pyCompressor.CandidateSubr(3, (0, 20))
        g2._global = True
        g3 = pyCompressor.CandidateSubr(3, (0, 30))
        g3._global = True
        l1 = pyCompressor.CandidateSubr(3, (0, 40))
        l1._global = False
        l2 = pyCompressor.CandidateSubr(3, (0, 50))
        l2._global = False
        l3 = pyCompressor.CandidateSubr(3, (0, 60))
        l3._global = False

        g1._encoding = [(3, l1)]
        g2._encoding = [(3, l2), (6, g3)]
        g3._encoding = []
        l1._encoding = []
        l2._encoding = [(3, l3)]
        l3._encoding = []

        lsubrs = self.empty_compreffor.collect_lsubrs_called_from([g1, g2, g3])
        self.assertSetEqual(lsubrs, {l1, l2, l3})

    def test_update_program_local(self):
        """Test update_program with only one replacement"""

        program = [7, 2, 10, 4, 8, 7, 0]
        substr = pyCompressor.CandidateSubr(3, (0, 1))
        substr._position = 5
        substr._fdidx = [0]
        substr._global = False
        encoding = [(1, substr)]
        bias = 0

        self.empty_compreffor.update_program(program, encoding, bias, [bias], 0)

        self.assertEqual(program, [7, 5, "callsubr", 8, 7, 0])

    def test_update_program_global(self):
        """Test update_program with only one replacement"""

        program = [7, 2, 10, 4, 8, 7, 0]
        substr = pyCompressor.CandidateSubr(3, (0, 1))
        substr._position = 5
        substr._fdidx = [0]
        substr._global = True
        encoding = [(1, substr)]
        bias = 0

        self.empty_compreffor.update_program(program, encoding, bias, [bias], 0)

        self.assertEqual(program, [7, 5, "callgsubr", 8, 7, 0])

    def test_update_program_multiple(self):
        """Test update_program with two replacements"""

        program = [7, 2, 10, 4, 8, 7, 0]
        substr = pyCompressor.CandidateSubr(3, (0, 1))
        substr._position = 5
        substr._global = True
        substr2 = pyCompressor.CandidateSubr(2, (0, 5))
        substr2._position = 21
        substr2._global = True
        encoding = [(1, substr), (5, substr2)]
        bias = 0

        self.empty_compreffor.update_program(program, encoding, bias, [bias], 0)

        self.assertEqual(program, [7, 5, "callgsubr", 8, 21, "callgsubr"])

    # TODO: make this test actually work
    def test_multiple_nested_subr_calls(self):
        """Test to make sure we can handle nested subrs. This is really just
        a case to make check we're encoding optimally."""

        glyph_set = {'a': (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 20),
                     'b': (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 21),
                     'c': (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 22),
                     'd': (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 23),
                     'e': (0, 1, 2, 3, 4, 5, 6, 7, 14, 15, 16, 17, 18, 19, 24),
                     'f': (0, 1, 2, 3, 4, 5, 6, 7, 14, 15, 16, 17, 18, 19, 25),
                     'g': (0, 1, 2, 3, 4, 5, 6, 7, 14, 15, 16, 17, 18, 19, 26),}
        glyph_set = DummyGlyphSet(glyph_set)

        ans = self.empty_compreffor.iterative_encode(glyph_set)
        print(ans["glyph_encodings"])
        print(ans["lsubrs"])
        print([s._encoding for s in ans["lsubrs"][0]])

    def test_expand_hintmask_single_middle(self):
        """Non-edge usage of expand_hintmask"""

        data = [1, 2, 3, 4, 5, ('hintmask', 7), 8, 9, 10]
        self.empty_compreffor.expand_hintmask(data)
        self.assertEqual(data, [1, 2, 3, 4, 5, 'hintmask', 7, 8, 9, 10])

    def test_expand_hintmask_multi_middle(self):
        """Non-edge usage of expand_hintmask with two items"""

        data = [1, ('hintmask', 3), 4, 5, ('hintmask', 7), 8, 9, 10]
        self.empty_compreffor.expand_hintmask(data)
        self.assertEqual(data, [1, 'hintmask', 3, 4, 5, 'hintmask', 7, 8, 9, 10])

    def test_expand_hintmask_multi_end(self):
        """Non-edge usage of expand_hintmask with two items, one at end"""

        data = [1, 2, 3, 4, 5, ('hintmask', 7), 8, ('hintmask', 10)]
        self.empty_compreffor.expand_hintmask(data)
        self.assertEqual(data, [1, 2, 3, 4, 5, 'hintmask', 7, 8, 'hintmask', 10])

    def test_collapse_hintmask_single_middle(self):
        """Non-edge usage of collapse_hintmask"""

        data = [1, 2, 3, 4, 5, 'hintmask', 7, 8, 9, 10]
        self.empty_compreffor.collapse_hintmask(data)
        self.assertEqual(data, [1, 2, 3, 4, 5, ('hintmask', 7), 8, 9, 10])

    def test_collapse_hintmask_multi_middle(self):
        """Non-edge usage of collapse_hintmask with two items"""

        data = [1, 'hintmask', 3, 4, 5, 'hintmask', 7, 8, 9, 10]
        self.empty_compreffor.collapse_hintmask(data)
        self.assertEqual(data, [1, ('hintmask', 3), 4, 5, ('hintmask', 7), 8, 9, 10])

    def test_collapse_hintmask_multi_end(self):
        """Non-edge usage of collapse_hintmask with two items, one at end"""

        data = [1, 2, 3, 4, 5, 'hintmask', 7, 8, 'hintmask', 10]
        self.empty_compreffor.collapse_hintmask(data)
        self.assertEqual(data, [1, 2, 3, 4, 5, ('hintmask', 7), 8, ('hintmask', 10)])

    def test_tokenCost(self):
        """Make sure single tokens can have their cost calculated"""

        tokenCost = pyCompressor.tokenCost

        self.assertEqual(tokenCost('hlineto'), 1)
        self.assertEqual(tokenCost('flex'), 2)
        self.assertEqual(tokenCost(107), 1)
        self.assertEqual(tokenCost(108), 2)

    def test_candidatesubr_len(self):
        """Make sure len returns the correct length"""

        self.assertEqual(len(self.cand_subr), 3)

    def test_candidatesubr_value(self):
        """Make sure the value is correct"""

        expected_value = (348, 374, 'rmoveto')

        self.assertEqual(self.cand_subr.value(), expected_value)



#!/usr/bin/env python
from setuptools import setup, find_packages, Extension
import os
from distutils.errors import DistutilsSetupError
from distutils import log
from distutils.dep_util import newer_group
import pkg_resources
import platform
import sys


needs_pytest = {'pytest', 'test'}.intersection(sys.argv)
pytest_runner = ['pytest_runner'] if needs_pytest else []
needs_wheel = {'bdist_wheel'}.intersection(sys.argv)
wheel = ['wheel'] if needs_wheel else []


# use Cython if available, else try use pre-generated .cpp sources
cython_min_version = '0.29.21'
try:
    pkg_resources.require("cython >= %s" % cython_min_version)
except pkg_resources.ResolutionError:
    with_cython = False
    print('Distribution mode: Compiling from Cython-generated .cpp sources.')
    from setuptools.command.build_ext import build_ext
else:
    with_cython = True
    print('Development mode: Compiling Cython modules from .pyx sources.')
    from Cython.Distutils import build_ext


class custom_build_ext(build_ext):
    """ Custom 'build_ext' command which allows to pass compiler-specific
    'extra_compile_args', 'define_macros' and 'undef_macros' options.
    """

    def build_extension(self, ext):
        sources = ext.sources
        if sources is None or not isinstance(sources, (list, tuple)):
            raise DistutilsSetupError(
                  "in 'ext_modules' option (extension '%s'), "
                  "'sources' must be present and must be "
                  "a list of source filenames" % ext.name)
        sources = list(sources)

        ext_path = self.get_ext_fullpath(ext.name)
        depends = sources + ext.depends
        if not (self.force or newer_group(depends, ext_path, 'newer')):
            log.debug("skipping '%s' extension (up-to-date)", ext.name)
            return
        else:
            log.info("building '%s' extension", ext.name)

        # do compiler specific customizations
        compiler_type = self.compiler.compiler_type

        if isinstance(ext.extra_compile_args, dict):
            extra_args_dict = ext.extra_compile_args or {}
            if compiler_type in extra_args_dict:
                extra_args = extra_args_dict[compiler_type]
            else:
                extra_args = extra_args_dict.get("default", [])
        else:
            extra_args = ext.extra_compile_args or []

        if isinstance(ext.define_macros, dict):
            macros_dict = ext.define_macros or {}
            if compiler_type in macros_dict:
                macros = macros_dict[compiler_type]
            else:
                macros = macros_dict.get("default", [])
        else:
            macros = ext.define_macros or []

        if isinstance(ext.undef_macros, dict):
            undef_macros_dict = ext.undef_macros
            for tp, undef in undef_macros_dict.items():
                if tp == compiler_type:
                    macros.append((undef,))
        else:
            for undef in ext.undef_macros:
                macros.append((undef,))

        # compile the source code to object files.
        objects = self.compiler.compile(sources,
                                        output_dir=self.build_temp,
                                        macros=macros,
                                        include_dirs=ext.include_dirs,
                                        debug=self.debug,
                                        extra_postargs=extra_args,
                                        depends=ext.depends)

        # Now link the object files together into a "shared object"
        if ext.extra_objects:
            objects.extend(ext.extra_objects)
        # TODO: do compiler-specific extra link args?
        extra_args = ext.extra_link_args or []

        # Detect target language, if not provided
        language = ext.language or self.compiler.detect_language(sources)

        self.compiler.link_shared_object(
            objects, ext_path,
            libraries=self.get_libraries(ext),
            library_dirs=ext.library_dirs,
            runtime_library_dirs=ext.runtime_library_dirs,
            extra_postargs=extra_args,
            export_symbols=self.get_export_symbols(ext),
            debug=self.debug,
            build_temp=self.build_temp,
            target_lang=language)


extensions = [
    Extension(
        "compreffor._compreffor",
        sources=[
            os.path.join('src', 'cython', (
                '_compreffor' + ('.pyx' if with_cython else '.cpp'))),
            os.path.join('src', 'cxx', "cffCompressor.cc"),
        ],
        depends=[os.path.join('src', 'cxx', 'cffCompressor.h')],
        extra_compile_args={
            "default": [
                "-std=c++0x", "-pthread",
                "-Wextra", "-Wno-unused", "-Wno-unused-parameter",
                # pass extra compiler flags on OS X to enable support for C++11
            ] + (["-stdlib=libc++", "-mmacosx-version-min=10.7"]
                 if platform.system() == "Darwin" else []),
            "msvc": ["/EHsc", "/Zi"],
        },
        language="c++",
    ),
]


with open('README.rst', 'r') as f:
    long_description = f.read()

setup_params = dict(
    name="compreffor",
    version="0.5.1",
    description="A CFF subroutinizer for fontTools.",
    long_description=long_description,
    author="Sam Fishman",
    license="Apache 2.0",
    package_dir={'': 'src/python'},
    packages=find_packages('src/python'),
    ext_modules=extensions,
    cmdclass={
        'build_ext': custom_build_ext,
    },
    setup_requires=pytest_runner + wheel,
    tests_require=[
        'pytest>=2.8',
    ],
    install_requires=[
        "fonttools>=4",
    ],
    python_requires=">=3.6",
    entry_points={
        'console_scripts': [
            "compreffor = compreffor.__main__:main",
        ]
    },
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Multimedia :: Graphics :: Graphics Conversion",
    ],
)


if __name__ == "__main__":
    setup(**setup_params)



#!/usr/bin/env python
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fontTools.ttLib import TTFont
from fontTools import cffLib
from fontTools.misc import psCharStrings
from fontTools.pens import basePen
import matplotlib.pyplot as plt
import functools
import itertools
import os
import argparse

"""
Prints out some stats about a set of fonts, mostly 
related to subroutines.

Dependencies:
  - matplotlib
  - fontTools

Usage:
>>> ./subr_grapher.py font1.otf font2.otf font3.otf cff_table.cff

NOTE: if the file extension is `cff`, it will be
interpreted as a raw CFF table.
"""

SINGLE_BYTE_OPS = set(['hstem',
                       'vstem',
                       'vmoveto',
                       'rlineto',
                       'hlineto',
                       'vlineto',
                       'rrcurveto',
                       'callsubr',
                       'return',
                       'endchar',
                       'blend',
                       'hstemhm',
                       'hintmask',
                       'cntrmask',
                       'rmoveto',
                       'hmoveto',
                       'vstemhm',
                       'rcurveline',
                       'rlinecurve',
                       'vvcurveto',
                       'hhcurveto',
                     # 'shortint',  # not really an operatr
                       'callgsubr',
                       'vhcurveto',
                       'hvcurveto'])

def tokenCost(token):
    """Calculate the bytecode size of a T2 Charstring token"""

    tp = type(token)
    if issubclass(tp, basestring):
        if token[:8] in ("hintmask", "cntrmask"):
            return 1 + len(token[9:])
        elif token in SINGLE_BYTE_OPS:
            return 1
        else:
            return 2
    elif tp == tuple:
        assert token[0] in ("hintmask", "cntrmask")
        return 1 + len(token[1])
    elif tp == int:
        if -107 <= token <= 107:
            return 1
        elif 108 <= token <= 1131 or -1131 <= token <= -108:
            return 2
        else:
            return 3
    elif tp == float:
        return 5
    assert 0

def get_cff(filename):
    if os.path.splitext(filename)[1] == '.cff':
        res = cffLib.CFFFontSet()
        res.decompile(open(filename), None)
        return res
    else:
        return TTFont(filename)['CFF '].cff

def get_cs_bytes(td, fds):
    count = 0
    for cs in td.GlobalSubrs:
        count += len(cs.bytecode)
    for fd in fds:
        try:
            for cs in fd.Private.Subrs:
                count += len(cs.bytecode)
        except AttributeError:
            pass
    for cs in td.CharStrings.values():
        count += len(cs.bytecode)
    return count

def print_n_subroutines(name, td, fds):
    print("%s:\n\tGlobal Subrs: %d" % (name, len(td.GlobalSubrs)))
    for i, fd in enumerate(fds):
        try:
            x = len(fd.Private.Subrs)
        except AttributeError:
            x = 0
        print("\tFD %d Subrs: %d" % (i, x))

def get_savings(td, fds):
    gsavings = [-(s.subr_cost + 2) if s.program else 0 for s in td.GlobalSubrs]
    lsavings = [[-(s.subr_cost + 2) if s.program else 0 for s in fd.Private.Subrs] for fd in fds]
    gusages = [0 for _ in td.GlobalSubrs]
    lusages = [[0 for _ in fd.Private.Subrs] for fd in fds]
    gbias = psCharStrings.calcSubrBias(td.GlobalSubrs)
    lbias = map(lambda fd: psCharStrings.calcSubrBias(fd.Private.Subrs)
                           if hasattr(fd.Private, 'Subrs') else 0,
                fds)
    
    def count_subr(idx, is_global, fdidx=-1):
        if is_global:
            gsavings[idx + gbias] += (td.GlobalSubrs[idx + gbias].subr_saving - tokenCost(idx) - 1)
            gusages[idx + gbias] += 1
            subr = td.GlobalSubrs[idx + gbias]
        else:
            assert fdidx >= 0
            lsavings[fdidx][idx + lbias[fdidx]] += (fds[fdidx].Private.Subrs[idx + lbias[fdidx]].subr_saving - tokenCost(idx) - 1)
            lusages[fdidx][idx + lbias[fdidx]] += 1
            subr = fds[fdidx].Private.Subrs[idx + lbias[fdidx]]

        # follow called subrs:
        for before, tok in zip(subr.program, subr.program[1:]):
            if tok == 'callgsubr':
                count_subr(before, True, fdidx)
            elif tok == 'callsubr':
                count_subr(before, False, fdidx)
    
    for g in td.charset:
        cs, sel = td.CharStrings.getItemAndSelector(g)
        for before, tok in zip(cs.program, cs.program[1:]):
            if tok == 'callgsubr':
                count_subr(before, True, sel)
            elif tok == 'callsubr':
                count_subr(before, False, sel)
                
    return ((gsavings, lsavings), (gusages, lusages))

def decompile_charstrings(td, fds):
    for cs in td.GlobalSubrs:
        cs.subr_cost = cs.subr_saving = len(cs.bytecode)
    for fd in fds:
        try:
            for cs in fd.Private.Subrs:
                cs.subr_cost = cs.subr_saving = len(cs.bytecode)
        except AttributeError:
            pass
    for g in td.charset:
        cs, sel = td.CharStrings.getItemAndSelector(g)
        cs.decompile()
    for cs in td.GlobalSubrs:
        if cs.program and cs.program[-1] == 'return':
            cs.subr_saving -= 1
    for fd in fds:
        try:
            for cs in fd.Private.Subrs:
                if cs.program and cs.program[-1] == 'return':
                    cs.subr_saving -= 1
        except AttributeError:
            pass

def get_raw_usages(td, fds):
    gusages = [0 for _ in td.GlobalSubrs]
    lusages = [[0 for _ in fd.Private.Subrs] for fd in fds]
    gbias = psCharStrings.calcSubrBias(td.GlobalSubrs)
    lbias = map(lambda fd: psCharStrings.calcSubrBias(fd.Private.Subrs)
                           if hasattr(fd.Private, 'Subrs') else 0,
                fds)
    gsels = [None for _ in td.GlobalSubrs]
    
    for g in td.charset:
        cs, sel = td.CharStrings.getItemAndSelector(g)
        for before, tok in zip(cs.program, cs.program[1:]):
            if tok == 'callgsubr':
                gusages[before + gbias] += 1
                gsels[before + gbias] = sel
            elif tok == 'callsubr':
                lusages[sel][before + lbias[sel]] += 1

    for cs, sel in zip(td.GlobalSubrs, gsels):
        for before, tok in zip(cs.program, cs.program[1:]):
            if tok == 'callgsubr':
                gusages[before + gbias] += 1
            elif tok == 'callsubr':
                lusages[sel][before + lbias[sel]] += 1

    for sel, fd in enumerate(fds):
        if hasattr(fd.Private, 'Subrs'):
            for cs in fd.Private.Subrs:
                for before, tok in zip(cs.program, cs.program[1:]):
                    if tok == 'callgsubr':
                        gusages[before + gbias] += 1
                    elif tok == 'callsubr':
                        lusages[sel][before + lbias[sel]] += 1

    return (gusages, lusages)

def main(filenames, show_graphs):
    names = map(os.path.basename, filenames)
    cffs = map(get_cff, filenames)
    tds = map(lambda f: f.topDictIndex[0], cffs)
    fds = map(lambda td: td.FDArray if hasattr(td, 'FDArray') else [], tds)

    n_bytes = map(get_cs_bytes, tds, fds)
    for name, b in zip(names, n_bytes):
        print("%s:\n\t%d bytes" % (name, b))

    map(decompile_charstrings, tds, fds)
    
    map(print_n_subroutines, names, tds, fds)

    sav_usag = map(get_savings, tds, fds)
    for name, (savings, usages) in zip(names, sav_usag):
        tot_savings = savings[0] + list(itertools.chain.from_iterable(savings[1]))
        tot_usages = usages[0] + list(itertools.chain.from_iterable(usages[1]))
        avg = float(sum(tot_savings)) / len(tot_savings)
        print("%s:\n\tAverage savings per subr: %f\n\tMax saving subr: %d\n\tMax usage subr: %d" % (name, avg, max(tot_savings), max(tot_usages)))

    if show_graphs:
        # plot subrs
        SHOW_START = 0
        SHOW_LEN = 200
        mins = []
        maxes = []
        plt.figure(0)
        for savings, usages in sav_usag:
            tot_savings = savings[0] + list(itertools.chain.from_iterable(savings[1]))
            plot_savings = sorted(tot_savings, reverse=True)[SHOW_START:SHOW_START+SHOW_LEN]
            plt.plot(range(len(plot_savings)), plot_savings)
            mins.append(min(plot_savings))
            maxes.append(max(plot_savings))
        plt.ylim([min(mins) - 1, max(maxes) + 1])
        plt.title("Subroutine Savings")
        plt.xlabel("Subroutine")
        plt.ylabel("Savings (bytes)")

        raw_usages = map(get_raw_usages, tds, fds)
        fig = 1
        for gusages, lusages in raw_usages:
            for idx, usages in zip(['Global'] + range(len(lusages)), [gusages] + lusages):
                if usages:
                    bias = psCharStrings.calcSubrBias(usages)
                    if bias == 1131:
                        orig_order_usages = usages[1024:1240] + usages[0:1024] + usages[1240:]
                    elif bias == 32768:
                        orig_order_usages = (usages[32661:32877] + usages[31637:32661] +
                                             usages[32877:33901] + usages[0:31637] + 
                                             usages[33901:])
                    else:
                        orig_order_usages = usages
                    plt.figure(fig)
                    plt.plot(range(len(orig_order_usages)), orig_order_usages, color='b')
                    plt.title("Subroutine usages for FD %s" % idx)
                    plt.axvline(215, 0, max(orig_order_usages), color='r')
                    plt.axvline(2263, 0, max(orig_order_usages), color='r')
                    plt.ylim([0, max(orig_order_usages)])
                    plt.xlim([0, len(orig_order_usages)])
                    fig += 1
        plt.show()

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
                        description="""FontTools Compreffor will take a CFF-flavored
                                       OpenType font and automatically detect
                                       repeated routines and generate subroutines
                                       to minimize the disk space needed to
                                       represent a font.""")
    parser.add_argument('filenames', help="the path to font files", nargs='+')
    parser.add_argument('-g', '--show-graphs', help="show graphs", action='store_true',
                        default=False)

    kwargs = vars(parser.parse_args())

    main(**kwargs)
